#include <gio/gio.h>

#if defined (__ELF__) && ( __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 6))
# define SECTION __attribute__ ((section (".gresource.gth"), aligned (8)))
#else
# define SECTION
#endif

static const SECTION union { const guint8 data[624]; const double alignment; void * const ptr;}  gth_resource_data = { {
  0x47, 0x56, 0x61, 0x72, 0x69, 0x61, 0x6e, 0x74, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x18, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x28, 0x08, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 
  0x02, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 
  0x03, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
  0x06, 0x00, 0x00, 0x00, 0x07, 0x00, 0x00, 0x00, 
  0x50, 0x79, 0xcc, 0x6a, 0x07, 0x00, 0x00, 0x00, 
  0x00, 0x01, 0x00, 0x00, 0x05, 0x00, 0x4c, 0x00, 
  0x08, 0x01, 0x00, 0x00, 0x0c, 0x01, 0x00, 0x00, 
  0xb0, 0xb7, 0x24, 0x30, 0x02, 0x00, 0x00, 0x00, 
  0x0c, 0x01, 0x00, 0x00, 0x06, 0x00, 0x4c, 0x00, 
  0x14, 0x01, 0x00, 0x00, 0x18, 0x01, 0x00, 0x00, 
  0x4b, 0x50, 0x90, 0x0b, 0x03, 0x00, 0x00, 0x00, 
  0x18, 0x01, 0x00, 0x00, 0x04, 0x00, 0x4c, 0x00, 
  0x1c, 0x01, 0x00, 0x00, 0x20, 0x01, 0x00, 0x00, 
  0xd4, 0xb5, 0x02, 0x00, 0xff, 0xff, 0xff, 0xff, 
  0x20, 0x01, 0x00, 0x00, 0x01, 0x00, 0x4c, 0x00, 
  0x24, 0x01, 0x00, 0x00, 0x28, 0x01, 0x00, 0x00, 
  0x6d, 0x89, 0x3b, 0x7c, 0x05, 0x00, 0x00, 0x00, 
  0x28, 0x01, 0x00, 0x00, 0x0e, 0x00, 0x76, 0x00, 
  0x38, 0x01, 0x00, 0x00, 0x47, 0x02, 0x00, 0x00, 
  0xbd, 0xb6, 0xd3, 0x49, 0x00, 0x00, 0x00, 0x00, 
  0x47, 0x02, 0x00, 0x00, 0x03, 0x00, 0x4c, 0x00, 
  0x4c, 0x02, 0x00, 0x00, 0x50, 0x02, 0x00, 0x00, 
  0x66, 0xf1, 0x4e, 0x60, 0x01, 0x00, 0x00, 0x00, 
  0x50, 0x02, 0x00, 0x00, 0x07, 0x00, 0x4c, 0x00, 
  0x58, 0x02, 0x00, 0x00, 0x5c, 0x02, 0x00, 0x00, 
  0xa7, 0x5b, 0x65, 0xf1, 0x06, 0x00, 0x00, 0x00, 
  0x5c, 0x02, 0x00, 0x00, 0x0d, 0x00, 0x4c, 0x00, 
  0x6c, 0x02, 0x00, 0x00, 0x70, 0x02, 0x00, 0x00, 
  0x64, 0x61, 0x74, 0x61, 0x2f, 0x00, 0x00, 0x00, 
  0x05, 0x00, 0x00, 0x00, 0x67, 0x6e, 0x6f, 0x6d, 
  0x65, 0x2f, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 
  0x6f, 0x72, 0x67, 0x2f, 0x01, 0x00, 0x00, 0x00, 
  0x2f, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 
  0x65, 0x78, 0x70, 0x6f, 0x72, 0x74, 0x2d, 0x6d, 
  0x65, 0x6e, 0x75, 0x2e, 0x75, 0x69, 0x00, 0x00, 
  0xdc, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 
  0x78, 0xda, 0x7d, 0x8e, 0x4b, 0x12, 0x82, 0x30, 
  0x10, 0x44, 0xf7, 0x9c, 0x62, 0xcc, 0xd6, 0x1a, 
  0x82, 0xe5, 0x36, 0xc0, 0x59, 0x14, 0x5a, 0x4d, 
  0x29, 0x89, 0x4e, 0x82, 0x72, 0x7c, 0x12, 0xfc, 
  0x94, 0x0b, 0xcb, 0xe5, 0xcc, 0x7b, 0x5d, 0xdd, 
  0xa6, 0x9d, 0x86, 0x0b, 0xdd, 0x21, 0xc1, 0x7a, 
  0x57, 0xab, 0x4d, 0x59, 0xa9, 0xb6, 0x29, 0x8c, 
  0x75, 0x11, 0x72, 0xd8, 0x75, 0x68, 0x0a, 0x22, 
  0xb3, 0x62, 0xa6, 0xcf, 0x87, 0x05, 0xb7, 0xd1, 
  0x0a, 0x02, 0x1d, 0xe3, 0x79, 0x4d, 0xdb, 0xb2, 
  0x22, 0xe6, 0x45, 0x1b, 0xe0, 0x46, 0xb2, 0x7d, 
  0xad, 0x30, 0x5d, 0xbd, 0x44, 0xce, 0xb7, 0xca, 
  0x24, 0xb1, 0x80, 0x2e, 0xa6, 0x86, 0x05, 0x3f, 
  0xb0, 0xe7, 0xa7, 0x92, 0x6a, 0xdf, 0x82, 0x7e, 
  0x19, 0x3f, 0x7c, 0x1f, 0x4f, 0x90, 0xff, 0x09, 
  0xa3, 0x73, 0x59, 0x1a, 0xae, 0xbf, 0x96, 0xcf, 
  0x72, 0x95, 0x44, 0x4f, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x28, 0x75, 0x75, 0x61, 0x79, 0x29, 0x75, 
  0x69, 0x2f, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 
  0x67, 0x54, 0x68, 0x75, 0x6d, 0x62, 0x2f, 0x00, 
  0x07, 0x00, 0x00, 0x00, 0x65, 0x78, 0x70, 0x6f, 
  0x72, 0x74, 0x5f, 0x74, 0x6f, 0x6f, 0x6c, 0x73, 
  0x2f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
} };

static GStaticResource static_resource = { gth_resource_data.data, sizeof (gth_resource_data.data), NULL, NULL, NULL };
extern GResource *gth_get_resource (void);
GResource *gth_get_resource (void)
{
  return g_static_resource_get_resource (&static_resource);
}
/*
  If G_HAS_CONSTRUCTORS is true then the compiler support *both* constructors and
  destructors, in a sane way, including e.g. on library unload. If not you're on
  your own.

  Some compilers need #pragma to handle this, which does not work with macros,
  so the way you need to use this is (for constructors):

  #ifdef G_DEFINE_CONSTRUCTOR_NEEDS_PRAGMA
  #pragma G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(my_constructor)
  #endif
  G_DEFINE_CONSTRUCTOR(my_constructor)
  static void my_constructor(void) {
   ...
  }

*/

#ifndef __GTK_DOC_IGNORE__

#if  __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 7)

#define G_HAS_CONSTRUCTORS 1

#define G_DEFINE_CONSTRUCTOR(_func) static void __attribute__((constructor)) _func (void);
#define G_DEFINE_DESTRUCTOR(_func) static void __attribute__((destructor)) _func (void);

#elif defined (_MSC_VER) && (_MSC_VER >= 1500)
/* Visual studio 2008 and later has _Pragma */

#define G_HAS_CONSTRUCTORS 1

#define G_DEFINE_CONSTRUCTOR(_func) \
  static void _func(void); \
  static int _func ## _wrapper(void) { _func(); return 0; } \
  __pragma(section(".CRT$XCU",read)) \
  __declspec(allocate(".CRT$XCU")) static int (* _array ## _func)(void) = _func ## _wrapper;

#define G_DEFINE_DESTRUCTOR(_func) \
  static void _func(void); \
  static int _func ## _constructor(void) { atexit (_func); return 0; } \
  __pragma(section(".CRT$XCU",read)) \
  __declspec(allocate(".CRT$XCU")) static int (* _array ## _func)(void) = _func ## _constructor;

#elif defined (_MSC_VER)

#define G_HAS_CONSTRUCTORS 1

/* Pre Visual studio 2008 must use #pragma section */
#define G_DEFINE_CONSTRUCTOR_NEEDS_PRAGMA 1
#define G_DEFINE_DESTRUCTOR_NEEDS_PRAGMA 1

#define G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(_func) \
  section(".CRT$XCU",read)
#define G_DEFINE_CONSTRUCTOR(_func) \
  static void _func(void); \
  static int _func ## _wrapper(void) { _func(); return 0; } \
  __declspec(allocate(".CRT$XCU")) static int (*p)(void) = _func ## _wrapper;

#define G_DEFINE_DESTRUCTOR_PRAGMA_ARGS(_func) \
  section(".CRT$XCU",read)
#define G_DEFINE_DESTRUCTOR(_func) \
  static void _func(void); \
  static int _func ## _constructor(void) { atexit (_func); return 0; } \
  __declspec(allocate(".CRT$XCU")) static int (* _array ## _func)(void) = _func ## _constructor;

#elif defined(__SUNPRO_C)

/* This is not tested, but i believe it should work, based on:
 * http://opensource.apple.com/source/OpenSSL098/OpenSSL098-35/src/fips/fips_premain.c
 */

#define G_HAS_CONSTRUCTORS 1

#define G_DEFINE_CONSTRUCTOR_NEEDS_PRAGMA 1
#define G_DEFINE_DESTRUCTOR_NEEDS_PRAGMA 1

#define G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(_func) \
  init(_func)
#define G_DEFINE_CONSTRUCTOR(_func) \
  static void _func(void);

#define G_DEFINE_DESTRUCTOR_PRAGMA_ARGS(_func) \
  fini(_func)
#define G_DEFINE_DESTRUCTOR(_func) \
  static void _func(void);

#else

/* constructors not supported for this compiler */

#endif

#endif /* __GTK_DOC_IGNORE__ */

#ifdef G_HAS_CONSTRUCTORS

#ifdef G_DEFINE_CONSTRUCTOR_NEEDS_PRAGMA
#pragma G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(resource_constructor)
#endif
G_DEFINE_CONSTRUCTOR(resource_constructor)
#ifdef G_DEFINE_DESTRUCTOR_NEEDS_PRAGMA
#pragma G_DEFINE_DESTRUCTOR_PRAGMA_ARGS(resource_destructor)
#endif
G_DEFINE_DESTRUCTOR(resource_destructor)

#else
#warning "Constructor not supported on this compiler, linking in resources will not work"
#endif

static void resource_constructor (void)
{
  g_static_resource_init (&static_resource);
}

static void resource_destructor (void)
{
  g_static_resource_fini (&static_resource);
}
